@@ -38,13 +38,17 @@ package resource_definitions
 
 import (
 	"crypto/sha256"
+	"encoding/json"
 	"fmt"
 	"math/rand"
 	"net/url"
+	"regexp"
+	"strconv"
 	"time"
 
 	operatorv1beta1 "github.com/cryostatio/cryostat-operator/api/v1beta1"
 	consolev1 "github.com/openshift/api/console/v1"
+	oauthv1 "github.com/openshift/api/oauth/v1"
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
 	rbacv1 "k8s.io/api/rbac/v1"
@@ -62,12 +66,13 @@ type ImageTags struct {
 	CoreImageTag       string
 	DatasourceImageTag string
 	GrafanaImageTag    string
+	ReportsImageTag    string
 }
 
 type ServiceSpecs struct {
 	CoreURL    *url.URL
-	CommandURL *url.URL
 	GrafanaURL *url.URL
+	ReportsURL *url.URL
 }
 
 // TLSConfig contains TLS-related information useful when creating other objects
@@ -76,10 +81,21 @@ type TLSConfig struct {
 	CryostatSecret string
 	// Name of the TLS secret for Grafana
 	GrafanaSecret string
+	// Name of the TLS secret for Reports Generator
+	ReportsSecret string
 	// Name of the secret containing the password for the keystore in CryostatSecret
 	KeystorePassSecret string
 }
 
+const (
+	cryostatHTTPContainerPort int32  = 8181
+	cryostatJMXContainerPort  int32  = 9091
+	grafanaContainerPort      int32  = 3000
+	datasourceContainerPort   int32  = 8080
+	reportsContainerPort      int32  = 10000
+	loopbackAddress           string = "127.0.0.1"
+)
+
 func NewPersistentVolumeClaimForCR(cr *operatorv1beta1.Cryostat) *corev1.PersistentVolumeClaim {
 	objMeta := metav1.ObjectMeta{
 		Name:      cr.Name,
@@ -130,6 +146,7 @@ func NewDeploymentForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, image
 			Labels: map[string]string{
 				"app":                    cr.Name,
 				"kind":                   "cryostat",
+				"component":              "cryostat",
 				"app.kubernetes.io/name": "cryostat",
 			},
 			Annotations: map[string]string{
@@ -139,8 +156,9 @@ func NewDeploymentForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, image
 		Spec: appsv1.DeploymentSpec{
 			Selector: &metav1.LabelSelector{
 				MatchLabels: map[string]string{
-					"app":  cr.Name,
-					"kind": "cryostat",
+					"app":       cr.Name,
+					"kind":      "cryostat",
+					"component": "cryostat",
 				},
 			},
 			Template: corev1.PodTemplateSpec{
@@ -148,8 +166,9 @@ func NewDeploymentForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, image
 					Name:      cr.Name,
 					Namespace: cr.Namespace,
 					Labels: map[string]string{
-						"app":  cr.Name,
-						"kind": "cryostat",
+						"app":       cr.Name,
+						"kind":      "cryostat",
+						"component": "cryostat",
 					},
 				},
 				Spec: *NewPodForCR(cr, specs, imageTags, tls, fsGroup, openshift),
@@ -158,6 +177,53 @@ func NewDeploymentForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, image
 	}
 }
 
+func NewDeploymentForReports(cr *operatorv1beta1.Cryostat, imageTags *ImageTags, tls *TLSConfig) *appsv1.Deployment {
+	if cr.Spec.ReportOptions == nil {
+		cr.Spec.ReportOptions = &operatorv1beta1.ReportConfiguration{Replicas: 0}
+	}
+	replicas := cr.Spec.ReportOptions.Replicas
+	return &appsv1.Deployment{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      cr.Name + "-reports",
+			Namespace: cr.Namespace,
+			Labels: map[string]string{
+				"app":                    cr.Name,
+				"kind":                   "cryostat",
+				"component":              "reports",
+				"app.kubernetes.io/name": "cryostat-reports",
+			},
+			Annotations: map[string]string{
+				"app.openshift.io/connects-to": cr.Name,
+			},
+		},
+		Spec: appsv1.DeploymentSpec{
+			Selector: &metav1.LabelSelector{
+				MatchLabels: map[string]string{
+					"app":       cr.Name,
+					"kind":      "cryostat",
+					"component": "reports",
+				},
+			},
+			Template: corev1.PodTemplateSpec{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      cr.Name + "-reports",
+					Namespace: cr.Namespace,
+					Labels: map[string]string{
+						"app":       cr.Name,
+						"kind":      "cryostat",
+						"component": "reports",
+					},
+				},
+				Spec: *NewPodForReports(cr, imageTags, tls),
+			},
+			Replicas: &replicas,
+		},
+	}
+}
+
+// Hostname alias for loopback address, to be used for health checks
+const healthCheckHostname = "cryostat-health.local"
+
 func NewPodForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTags *ImageTags,
 	tls *TLSConfig, fsGroup int64, openshift bool) *corev1.PodSpec {
 	var containers []corev1.Container
@@ -172,69 +238,53 @@ func NewPodForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTags *I
 			NewJfrDatasourceContainer(cr, imageTags.DatasourceImageTag),
 		}
 	}
-	volumes := []corev1.Volume{
-		{
-			Name: cr.Name,
-			VolumeSource: corev1.VolumeSource{
-				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
-					ClaimName: cr.Name,
-				},
-			},
-		},
-	}
-	readOnlyMode := int32(0440)
-	if tls != nil {
-		// Create certificate secret volumes in deployment
-		volSources := []corev1.VolumeProjection{
-			{
-				// Add Cryostat self-signed CA
-				Secret: &corev1.SecretProjection{
-					LocalObjectReference: corev1.LocalObjectReference{
-						Name: tls.CryostatSecret,
-					},
-					Items: []corev1.KeyToPath{
-						{
-							Key:  CAKey,
-							Path: cr.Name + "-ca.crt",
-							Mode: &readOnlyMode,
-						},
-					},
-				},
-			},
-		}
 
-		// Add any TrustedCertSecrets as volumes
-		for _, secret := range cr.Spec.TrustedCertSecrets {
-			var key string
-			if secret.CertificateKey != nil {
-				key = *secret.CertificateKey
-			} else {
-				key = operatorv1beta1.DefaultCertificateKey
-			}
-			source := corev1.VolumeProjection{
-				Secret: &corev1.SecretProjection{
-					LocalObjectReference: corev1.LocalObjectReference{
-						Name: secret.SecretName,
-					},
-					Items: []corev1.KeyToPath{
-						{
-							Key:  key,
-							Path: fmt.Sprintf("%s_%s", secret.SecretName, key),
-							Mode: &readOnlyMode,
-						},
-					},
-				},
-			}
-			volSources = append(volSources, source)
+	volumes := newVolumeForCR(cr)
+	volSources := []corev1.VolumeProjection{}
+	readOnlyMode := int32(0440)
+
+	// Add any TrustedCertSecrets as volumes
+	for _, secret := range cr.Spec.TrustedCertSecrets {
+		var key string
+		if secret.CertificateKey != nil {
+			key = *secret.CertificateKey
+		} else {
+			key = operatorv1beta1.DefaultCertificateKey
 		}
-		certVolume := corev1.Volume{
-			Name: "cert-secrets",
-			VolumeSource: corev1.VolumeSource{
-				Projected: &corev1.ProjectedVolumeSource{
-					Sources: volSources,
+		source := corev1.VolumeProjection{
+			Secret: &corev1.SecretProjection{
+				LocalObjectReference: corev1.LocalObjectReference{
+					Name: secret.SecretName,
+				},
+				Items: []corev1.KeyToPath{
+					{
+						Key:  key,
+						Path: fmt.Sprintf("%s_%s", secret.SecretName, key),
+						Mode: &readOnlyMode,
+					},
 				},
 			},
 		}
+		volSources = append(volSources, source)
+	}
+
+	if tls != nil {
+		volSources = append(volSources, corev1.VolumeProjection{
+			// Add Cryostat self-signed CA
+			Secret: &corev1.SecretProjection{
+				LocalObjectReference: corev1.LocalObjectReference{
+					Name: tls.CryostatSecret,
+				},
+				Items: []corev1.KeyToPath{
+					{
+						Key:  CAKey,
+						Path: cr.Name + "-ca.crt",
+						Mode: &readOnlyMode,
+					},
+				},
+			},
+		})
+
 		keyVolume := corev1.Volume{
 			Name: "keystore",
 			VolumeSource: corev1.VolumeSource{
@@ -251,7 +301,7 @@ func NewPodForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTags *I
 			},
 		}
 
-		volumes = append(volumes, certVolume, keyVolume)
+		volumes = append(volumes, keyVolume)
 
 		if !cr.Spec.Minimal {
 			grafanaSecretVolume := corev1.Volume{
@@ -266,6 +316,17 @@ func NewPodForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTags *I
 		}
 	}
 
+	// Project certificate secrets into deployment
+	certVolume := corev1.Volume{
+		Name: "cert-secrets",
+		VolumeSource: corev1.VolumeSource{
+			Projected: &corev1.ProjectedVolumeSource{
+				Sources: volSources,
+			},
+		},
+	}
+	volumes = append(volumes, certVolume)
+
 	// Add any EventTemplates as volumes
 	for _, template := range cr.Spec.EventTemplates {
 		eventTemplateVolume := corev1.Volume{
@@ -292,11 +353,139 @@ func NewPodForCR(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTags *I
 	sc := &corev1.PodSecurityContext{
 		FSGroup: &fsGroup,
 	}
+
+	// Use HostAlias for loopback address to allow health checks to
+	// work over HTTPS with hostname added as a SubjectAltName
+	hostAliases := []corev1.HostAlias{
+		{
+			IP: loopbackAddress,
+			Hostnames: []string{
+				healthCheckHostname,
+			},
+		},
+	}
 	return &corev1.PodSpec{
 		ServiceAccountName: cr.Name,
 		Volumes:            volumes,
 		Containers:         containers,
 		SecurityContext:    sc,
+		HostAliases:        hostAliases,
+	}
+}
+
+func NewPodForReports(cr *operatorv1beta1.Cryostat, imageTags *ImageTags, tls *TLSConfig) *corev1.PodSpec {
+	resources := corev1.ResourceRequirements{}
+	if cr.Spec.ReportOptions != nil {
+		resources = cr.Spec.ReportOptions.Resources
+	}
+	cpus := int64(1)
+	if requests := resources.Requests; requests != nil {
+		if cpu := requests.Cpu(); cpu != nil {
+			cpus = cpu.Value()
+		}
+	}
+	if limits := resources.Limits; limits != nil {
+		if cpu := limits.Cpu(); cpu != nil {
+			cpus = cpu.Value()
+		}
+	}
+	javaOpts := fmt.Sprintf("-XX:+PrintCommandLineFlags -XX:ActiveProcessorCount=%d -Dorg.openjdk.jmc.flightrecorder.parser.singlethreaded=%t", cpus, cpus < 2)
+
+	envs := []corev1.EnvVar{
+		{
+			Name:  "QUARKUS_HTTP_HOST",
+			Value: "0.0.0.0",
+		},
+		{
+			Name:  "JAVA_OPTIONS",
+			Value: javaOpts,
+		},
+	}
+	mounts := []corev1.VolumeMount{}
+	volumes := []corev1.Volume{}
+
+	// Configure TLS key/cert if enabled
+	livenessProbeScheme := corev1.URISchemeHTTP
+	if tls != nil {
+		tlsEnvs := []corev1.EnvVar{
+			{
+				Name:  "QUARKUS_HTTP_SSL_PORT",
+				Value: strconv.Itoa(int(reportsContainerPort)),
+			},
+			{
+				Name:  "QUARKUS_HTTP_SSL_CERTIFICATE_KEY_FILE",
+				Value: fmt.Sprintf("/var/run/secrets/operator.cryostat.io/%s/%s", tls.ReportsSecret, corev1.TLSPrivateKeyKey),
+			},
+			{
+				Name:  "QUARKUS_HTTP_SSL_CERTIFICATE_FILE",
+				Value: fmt.Sprintf("/var/run/secrets/operator.cryostat.io/%s/%s", tls.ReportsSecret, corev1.TLSCertKey),
+			},
+			{
+				Name:  "QUARKUS_HTTP_INSECURE_REQUESTS",
+				Value: "disabled",
+			},
+		}
+
+		tlsSecretMount := corev1.VolumeMount{
+			Name:      "reports-tls-secret",
+			MountPath: "/var/run/secrets/operator.cryostat.io/" + tls.ReportsSecret,
+			ReadOnly:  true,
+		}
+
+		secretVolume := corev1.Volume{
+			Name: "reports-tls-secret",
+			VolumeSource: corev1.VolumeSource{
+				Secret: &corev1.SecretVolumeSource{
+					SecretName: tls.ReportsSecret,
+				},
+			},
+		}
+
+		envs = append(envs, tlsEnvs...)
+		mounts = append(mounts, tlsSecretMount)
+		volumes = append(volumes, secretVolume)
+
+		// Use HTTPS for liveness probe
+		livenessProbeScheme = corev1.URISchemeHTTPS
+
+	} else {
+		envs = append(envs, corev1.EnvVar{
+			Name:  "QUARKUS_HTTP_PORT",
+			Value: strconv.Itoa(int(reportsContainerPort)),
+		})
+	}
+
+	probeHandler := corev1.Handler{
+		HTTPGet: &corev1.HTTPGetAction{
+			Scheme: livenessProbeScheme,
+			Port:   intstr.IntOrString{IntVal: reportsContainerPort},
+			Path:   "/health",
+		},
+	}
+	return &corev1.PodSpec{
+		ServiceAccountName: cr.Name,
+		Containers: []corev1.Container{
+			{
+				Name:            cr.Name + "-reports",
+				Image:           imageTags.ReportsImageTag,
+				ImagePullPolicy: getPullPolicy(imageTags.ReportsImageTag),
+				Ports: []corev1.ContainerPort{
+					{
+						ContainerPort: reportsContainerPort,
+					},
+				},
+				Env:          envs,
+				VolumeMounts: mounts,
+				Resources:    resources,
+				LivenessProbe: &corev1.Probe{
+					Handler: probeHandler,
+				},
+				StartupProbe: &corev1.Probe{
+					Handler: probeHandler,
+				},
+			},
+		},
+		Volumes: volumes,
 	}
 }
 
@@ -306,22 +495,11 @@ func NewCoreContainer(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTa
 	archivePath := "/opt/cryostat.d/recordings.d"
 	templatesPath := "/opt/cryostat.d/templates.d"
 	clientlibPath := "/opt/cryostat.d/clientlib.d"
+	probesPath := "/opt/cryostat.d/probes.d"
 	envs := []corev1.EnvVar{
-		{
-			Name:  "CRYOSTAT_SSL_PROXIED",
-			Value: "true",
-		},
-		{
-			Name:  "CRYOSTAT_ALLOW_UNTRUSTED_SSL",
-			Value: "true",
-		},
 		{
 			Name:  "CRYOSTAT_WEB_PORT",
-			Value: "8181",
-		},
-		{
-			Name:  "CRYOSTAT_LISTEN_PORT",
-			Value: "9090",
+			Value: strconv.Itoa(int(cryostatHTTPContainerPort)),
 		},
 		{
 			Name:  "CRYOSTAT_CONFIG_PATH",
@@ -339,6 +517,14 @@ func NewCoreContainer(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTa
 			Name:  "CRYOSTAT_CLIENTLIB_PATH",
 			Value: clientlibPath,
 		},
+		{
+			Name:  "CRYOSTAT_PROBE_TEMPLATE_PATH",
+			Value: probesPath,
+		},
+		{
+			Name:  "CRYOSTAT_ENABLE_JDP_BROADCAST",
+			Value: "false",
+		},
 	}
 	if specs.CoreURL != nil {
 		coreEnvs := []corev1.EnvVar{
@@ -353,19 +539,63 @@ func NewCoreContainer(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTa
 		}
 		envs = append(envs, coreEnvs...)
 	}
-	if specs.CommandURL != nil {
-		commandEnvs := []corev1.EnvVar{
+	if specs.ReportsURL != nil {
+		reportsEnvs := []corev1.EnvVar{
 			{
-				Name:  "CRYOSTAT_EXT_LISTEN_PORT",
-				Value: getPort(specs.CommandURL),
-			},
-			{
-				Name:  "CRYOSTAT_LISTEN_HOST",
-				Value: specs.CommandURL.Hostname(),
+				Name:  "CRYOSTAT_REPORT_GENERATOR",
+				Value: specs.ReportsURL.String(),
 			},
 		}
-		envs = append(envs, commandEnvs...)
+		envs = append(envs, reportsEnvs...)
+	} else {
+		subProcessMaxHeapSize := "200"
+		if cr.Spec.ReportOptions.SubProcessMaxHeapSize != 0 {
+			subProcessMaxHeapSize = strconv.Itoa(int(cr.Spec.ReportOptions.SubProcessMaxHeapSize))
+		}
+		subprocessReportHeapEnv := []corev1.EnvVar{
+			{
+				Name:  "CRYOSTAT_REPORT_GENERATION_MAX_HEAP",
+				Value: subProcessMaxHeapSize,
+			},
+		}
+		envs = append(envs, subprocessReportHeapEnv...)
 	}
+
+	if cr.Spec.MaxWsConnections != 0 {
+		maxWsConnections := strconv.Itoa(int(cr.Spec.MaxWsConnections))
+		maxWsConnectionsEnv := []corev1.EnvVar{
+			{
+				Name:  "CRYOSTAT_MAX_WS_CONNECTIONS",
+				Value: maxWsConnections,
+			},
+		}
+		envs = append(envs, maxWsConnectionsEnv...)
+	}
+
+	targetCacheSize := "-1"
+	targetCacheTTL := "10"
+	if cr.Spec.JmxCacheOptions != nil {
+
+		if cr.Spec.JmxCacheOptions.TargetCacheSize != 0 {
+			targetCacheSize = strconv.Itoa(int(cr.Spec.JmxCacheOptions.TargetCacheSize))
+		}
+
+		if cr.Spec.JmxCacheOptions.TargetCacheTTL != 0 {
+			targetCacheTTL = strconv.Itoa(int(cr.Spec.JmxCacheOptions.TargetCacheTTL))
+		}
+	}
+	jmxCacheEnvs := []corev1.EnvVar{
+		{
+			Name:  "CRYOSTAT_TARGET_CACHE_SIZE",
+			Value: targetCacheSize,
+		},
+		{
+			Name:  "CRYOSTAT_TARGET_CACHE_TTL",
+			Value: targetCacheTTL,
+		},
+	}
+	envs = append(envs, jmxCacheEnvs...)
+
 	if openshift {
 		// Force OpenShift platform strategy
 		openshiftEnvs := []corev1.EnvVar{
@@ -375,7 +605,15 @@ func NewCoreContainer(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTa
 			},
 			{
 				Name:  "CRYOSTAT_AUTH_MANAGER",
-				Value: "io.cryostat.net.OpenShiftAuthManager",
+				Value: "io.cryostat.net.openshift.OpenShiftAuthManager",
+			},
+			{
+				Name:  "CRYOSTAT_OAUTH_CLIENT_ID",
+				Value: cr.Name,
+			},
+			{
+				Name:  "CRYOSTAT_OAUTH_ROLE",
+				Value: "cryostat-operator-oauth-client",
 			},
 		}
 		envs = append(envs, openshiftEnvs...)
@@ -411,25 +649,41 @@ func NewCoreContainer(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTa
 			MountPath: clientlibPath,
 			SubPath:   "clientlib",
 		},
+		{
+			Name:      cr.Name,
+			MountPath: probesPath,
+			SubPath:   "probes",
+		},
 		{
 			Name:      cr.Name,
 			MountPath: "truststore",
 			SubPath:   "truststore",
 		},
+		{
+			// Mount the CA cert and user certificates in the expected /truststore location
+			Name:      "cert-secrets",
+			MountPath: "/truststore/operator",
+			ReadOnly:  true,
+		},
 	}
 
 	if !cr.Spec.Minimal {
 		grafanaVars := []corev1.EnvVar{
 			{
 				Name:  "GRAFANA_DATASOURCE_URL",
-				Value: DatasourceURL,
+				Value: datasourceURL,
 			},
 		}
 		if specs.GrafanaURL != nil {
-			grafanaVars = append(grafanaVars, corev1.EnvVar{
-				Name:  "GRAFANA_DASHBOARD_URL",
-				Value: specs.GrafanaURL.String(),
-			})
+			grafanaVars = append(grafanaVars,
+				corev1.EnvVar{
+					Name:  "GRAFANA_DASHBOARD_EXT_URL",
+					Value: specs.GrafanaURL.String(),
+				},
+				corev1.EnvVar{
+					Name:  "GRAFANA_DASHBOARD_URL",
+					Value: getInternalDashboardURL(tls),
+				})
 		}
 		envs = append(envs, grafanaVars...)
 	}
@@ -441,6 +695,13 @@ func NewCoreContainer(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTa
 			Name:  "CRYOSTAT_DISABLE_SSL",
 			Value: "true",
 		})
+		// Set CRYOSTAT_SSL_PROXIED if Ingress/Route use HTTPS
+		if specs.CoreURL != nil && specs.CoreURL.Scheme == "https" {
+			envs = append(envs, corev1.EnvVar{
+				Name:  "CRYOSTAT_SSL_PROXIED",
+				Value: "true",
+			})
+		}
 	} else {
 		// Configure keystore location and password in expected environment variables
 		envs = append(envs, corev1.EnvVar{
@@ -462,14 +723,7 @@ func NewCoreContainer(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTa
 			ReadOnly:  true,
 		}
 
-		// Mount the CA cert and user certificates in the expected /truststore location
-		caCertMount := corev1.VolumeMount{
-			Name:      "cert-secrets",
-			MountPath: "/truststore/operator",
-			ReadOnly:  true,
-		}
-
-		mounts = append(mounts, keystoreMount, caCertMount)
+		mounts = append(mounts, keystoreMount)
 
 		// Use HTTPS for liveness probe
 		livenessProbeScheme = corev1.URISchemeHTTPS
@@ -488,28 +742,27 @@ func NewCoreContainer(cr *operatorv1beta1.Cryostat, specs *ServiceSpecs, imageTa
 
 	probeHandler := corev1.Handler{
 		HTTPGet: &corev1.HTTPGetAction{
-			Port:   intstr.IntOrString{IntVal: 8181},
+			Port:   intstr.IntOrString{IntVal: cryostatHTTPContainerPort},
 			Path:   "/health",
 			Scheme: livenessProbeScheme,
 		},
 	}
 	return corev1.Container{
-		Name:         cr.Name,
-		Image:        imageTag,
-		VolumeMounts: mounts,
+		Name:            cr.Name,
+		Image:           imageTag,
+		ImagePullPolicy: getPullPolicy(imageTag),
+		VolumeMounts:    mounts,
 		Ports: []corev1.ContainerPort{
 			{
-				ContainerPort: 8181,
+				ContainerPort: cryostatHTTPContainerPort,
 			},
 			{
-				ContainerPort: 9090,
-			},
-			{
-				ContainerPort: 9091,
+				ContainerPort: cryostatJMXContainerPort,
 			},
 		},
-		Env:     envs,
-		EnvFrom: envsFrom,
+		Env:       envs,
+		EnvFrom:   envsFrom,
+		Resources: cr.Spec.Resources.CoreResources,
 		LivenessProbe: &corev1.Probe{
 			Handler: probeHandler,
 		},
@@ -548,7 +801,7 @@ func NewGrafanaContainer(cr *operatorv1beta1.Cryostat, imageTag string, tls *TLS
 	envs := []corev1.EnvVar{
 		{
 			Name:  "JFR_DATASOURCE_URL",
-			Value: DatasourceURL,
+			Value: datasourceURL,
 		},
 	}
 	mounts := []corev1.VolumeMount{}
@@ -584,12 +837,13 @@ func NewGrafanaContainer(cr *operatorv1beta1.Cryostat, imageTag string, tls *TLS
 		livenessProbeScheme = corev1.URISchemeHTTPS
 	}
 	return corev1.Container{
-		Name:         cr.Name + "-grafana",
-		Image:        imageTag,
-		VolumeMounts: mounts,
+		Name:            cr.Name + "-grafana",
+		Image:           imageTag,
+		ImagePullPolicy: getPullPolicy(imageTag),
+		VolumeMounts:    mounts,
 		Ports: []corev1.ContainerPort{
 			{
-				ContainerPort: 3000,
+				ContainerPort: grafanaContainerPort,
 			},
 		},
 		Env: envs,
@@ -611,92 +865,86 @@ func NewGrafanaContainer(cr *operatorv1beta1.Cryostat, imageTag string, tls *TLS
 				},
 			},
 		},
+		Resources: cr.Spec.Resources.GrafanaResources,
 	}
 }
 
-const datasourceHost = "127.0.0.1"
-const datasourcePort = "8080"
-
-// DatasourceURL contains the fixed URL to jfr-datasource's web server
-const DatasourceURL = "http://" + datasourceHost + ":" + datasourcePort
+// datasourceURL contains the fixed URL to jfr-datasource's web server
+var datasourceURL = "http://" + loopbackAddress + ":" + strconv.Itoa(int(datasourceContainerPort))
 
 func NewJfrDatasourceContainer(cr *operatorv1beta1.Cryostat, imageTag string) corev1.Container {
 	return corev1.Container{
-		Name:  cr.Name + "-jfr-datasource",
-		Image: imageTag,
+		Name:            cr.Name + "-jfr-datasource",
+		Image:           imageTag,
+		ImagePullPolicy: getPullPolicy(imageTag),
 		Ports: []corev1.ContainerPort{
 			{
-				ContainerPort: 8080,
+				ContainerPort: datasourceContainerPort,
 			},
 		},
 		Env: []corev1.EnvVar{
 			{
 				Name:  "LISTEN_HOST",
-				Value: datasourceHost,
+				Value: loopbackAddress,
 			},
 		},
 		// Can't use HTTP probe since the port is not exposed over the network
 		LivenessProbe: &corev1.Probe{
 			Handler: corev1.Handler{
 				Exec: &corev1.ExecAction{
-					Command: []string{"curl", "--fail", DatasourceURL},
+					Command: []string{"curl", "--fail", datasourceURL},
 				},
 			},
 		},
+		Resources: cr.Spec.Resources.DataSourceResources,
 	}
 }
 
-func NewExporterService(cr *operatorv1beta1.Cryostat) *corev1.Service {
+func NewCoreService(cr *operatorv1beta1.Cryostat) *corev1.Service {
+	// Check CR for config
+	var config *operatorv1beta1.CoreServiceConfig
+	if cr.Spec.ServiceOptions == nil || cr.Spec.ServiceOptions.CoreConfig == nil {
+		config = &operatorv1beta1.CoreServiceConfig{}
+	} else {
+		config = cr.Spec.ServiceOptions.CoreConfig
+	}
+
+	// Apply common service defaults
+	configureService(&config.ServiceConfig, cr.Name, "cryostat")
+
+	// Apply default HTTP and JMX port if not provided
+	if config.HTTPPort == nil {
+		httpPort := cryostatHTTPContainerPort
+		config.HTTPPort = &httpPort
+	}
+	if config.JMXPort == nil {
+		jmxPort := cryostatJMXContainerPort
+		config.JMXPort = &jmxPort
+	}
+
 	return &corev1.Service{
 		ObjectMeta: metav1.ObjectMeta{
-			Name:      cr.Name,
-			Namespace: cr.Namespace,
-			Labels: map[string]string{
+			Name:        cr.Name,
+			Namespace:   cr.Namespace,
+			Labels:      config.Labels,
+			Annotations: config.Annotations,
+		},
+		Spec: corev1.ServiceSpec{
+			Type: *config.ServiceType,
+			Selector: map[string]string{
 				"app":       cr.Name,
 				"component": "cryostat",
 			},
-		},
-		Spec: corev1.ServiceSpec{
-			Type: corev1.ServiceTypeLoadBalancer,
-			Selector: map[string]string{
-				"app": cr.Name,
-			},
 			Ports: []corev1.ServicePort{
 				{
-					Name:       "export",
-					Port:       8181,
-					TargetPort: intstr.IntOrString{IntVal: 8181},
+					Name:       "http",
+					Port:       *config.HTTPPort,
+					TargetPort: intstr.IntOrString{IntVal: cryostatHTTPContainerPort},
 				},
 				{
 					Name:       "jfr-jmx",
-					Port:       9091,
-					TargetPort: intstr.IntOrString{IntVal: 9091},
-				},
-			},
-		},
-	}
-}
-
-func NewCommandChannelService(cr *operatorv1beta1.Cryostat) *corev1.Service {
-	return &corev1.Service{
-		ObjectMeta: metav1.ObjectMeta{
-			Name:      cr.Name + "-command",
-			Namespace: cr.Namespace,
-			Labels: map[string]string{
-				"app":       cr.Name,
-				"component": "command-channel",
-			},
-		},
-		Spec: corev1.ServiceSpec{
-			Type: corev1.ServiceTypeLoadBalancer,
-			Selector: map[string]string{
-				"app": cr.Name,
-			},
-			Ports: []corev1.ServicePort{
-				{
-					Name:       "cmdchan",
-					Port:       9090,
-					TargetPort: intstr.IntOrString{IntVal: 9090},
+					Port:       *config.JMXPort,
+					TargetPort: intstr.IntOrString{IntVal: cryostatJMXContainerPort},
 				},
 			},
 		},
@@ -704,24 +952,24 @@ func NewCommandChannelService(cr *operatorv1beta1.Cryostat) *corev1.Service {
 }
 
 func NewGrafanaService(cr *operatorv1beta1.Cryostat) *corev1.Service {
+	config := getGrafanaServiceConfig(cr)
 	return &corev1.Service{
 		ObjectMeta: metav1.ObjectMeta{
-			Name:      cr.Name + "-grafana",
-			Namespace: cr.Namespace,
-			Labels: map[string]string{
-				"app":       cr.Name,
-				"component": "grafana",
-			},
+			Name:        cr.Name + "-grafana",
+			Namespace:   cr.Namespace,
+			Labels:      config.Labels,
+			Annotations: config.Annotations,
 		},
 		Spec: corev1.ServiceSpec{
-			Type: corev1.ServiceTypeLoadBalancer,
+			Type: *config.ServiceType,
 			Selector: map[string]string{
-				"app": cr.Name,
+				"app":       cr.Name,
+				"component": "cryostat",
 			},
 			Ports: []corev1.ServicePort{
 				{
-					Name:       "3000-tcp",
-					Port:       3000,
+					Name:       "http",
+					Port:       *config.HTTPPort,
 					TargetPort: intstr.IntOrString{IntVal: 3000},
 				},
 			},
@@ -729,6 +977,86 @@ func NewGrafanaService(cr *operatorv1beta1.Cryostat) *corev1.Service {
 	}
 }
 
+func NewReportService(cr *operatorv1beta1.Cryostat) *corev1.Service {
+	// Check CR for config
+	var config *operatorv1beta1.ReportsServiceConfig
+	if cr.Spec.ServiceOptions == nil || cr.Spec.ServiceOptions.ReportsConfig == nil {
+		config = &operatorv1beta1.ReportsServiceConfig{}
+	} else {
+		config = cr.Spec.ServiceOptions.ReportsConfig
+	}
+
+	// Apply common service defaults
+	configureService(&config.ServiceConfig, cr.Name, "reports")
+
+	// Apply default HTTP port if not provided
+	if config.HTTPPort == nil {
+		httpPort := reportsContainerPort
+		config.HTTPPort = &httpPort
+	}
+
+	return &corev1.Service{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:        cr.Name + "-reports",
+			Namespace:   cr.Namespace,
+			Labels:      config.Labels,
+			Annotations: config.Annotations,
+		},
+		Spec: corev1.ServiceSpec{
+			Type: *config.ServiceType,
+			Selector: map[string]string{
+				"app":       cr.Name,
+				"component": "reports",
+			},
+			Ports: []corev1.ServicePort{
+				{
+					Name:       "http",
+					Port:       *config.HTTPPort,
+					TargetPort: intstr.IntOrString{IntVal: reportsContainerPort},
+				},
+			},
+		},
+	}
+}
+
+func getGrafanaServiceConfig(cr *operatorv1beta1.Cryostat) *operatorv1beta1.GrafanaServiceConfig {
+	// Check CR for config
+	var config *operatorv1beta1.GrafanaServiceConfig
+	if cr.Spec.ServiceOptions == nil || cr.Spec.ServiceOptions.GrafanaConfig == nil {
+		config = &operatorv1beta1.GrafanaServiceConfig{}
+	} else {
+		config = cr.Spec.ServiceOptions.GrafanaConfig
+	}
+
+	// Apply common service defaults
+	configureService(&config.ServiceConfig, cr.Name, "cryostat")
+
+	// Apply default HTTP port if not provided
+	if config.HTTPPort == nil {
+		httpPort := grafanaContainerPort
+		config.HTTPPort = &httpPort
+	}
+
+	return config
+}
+
+func configureService(config *operatorv1beta1.ServiceConfig, appLabel string, componentLabel string) {
+	if config.ServiceType == nil {
+		svcType := corev1.ServiceTypeClusterIP
+		config.ServiceType = &svcType
+	}
+	if config.Labels == nil {
+		config.Labels = map[string]string{}
+	}
+	if config.Annotations == nil {
+		config.Annotations = map[string]string{}
+	}
+
+	// Add required labels, overriding any user-specified labels with the same keys
+	config.Labels["app"] = appLabel
+	config.Labels["component"] = componentLabel
+}
+
 // JMXSecretNameSuffix is the suffix to be appended to the name of a
 // Cryostat CR to name its JMX credentials secret
 const JMXSecretNameSuffix = "-jmx-auth"
@@ -764,13 +1092,37 @@ func NewKeystoreSecretForCR(cr *operatorv1beta1.Cryostat) *corev1.Secret {
 	}
 }
 
-func NewServiceAccountForCR(cr *operatorv1beta1.Cryostat) *corev1.ServiceAccount {
+func NewServiceAccountForCR(cr *operatorv1beta1.Cryostat, isOpenShift bool) (*corev1.ServiceAccount, error) {
+	annotations := make(map[string]string)
+
+	if isOpenShift {
+		OAuthRedirectReference := &oauthv1.OAuthRedirectReference{
+			Reference: oauthv1.RedirectReference{
+				Kind: "Route",
+				Name: cr.Name,
+			},
+		}
+
+		ref, err := json.Marshal(OAuthRedirectReference)
+		if err != nil {
+			return nil, err
+		}
+
+		annotations = map[string]string{
+			"serviceaccounts.openshift.io/oauth-redirectreference.route": string(ref),
+		}
+	}
+
 	return &corev1.ServiceAccount{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      cr.Name,
 			Namespace: cr.Namespace,
+			Labels: map[string]string{
+				"app": "cryostat",
+			},
+			Annotations: annotations,
 		},
-	}
+	}, nil
 }
 
 func NewRoleForCR(cr *operatorv1beta1.Cryostat) *rbacv1.Role {
@@ -882,9 +1234,63 @@ func getPort(url *url.URL) string {
 	return "80"
 }
 
+func getInternalDashboardURL(tls *TLSConfig) string {
+	scheme := "https"
+	if tls == nil {
+		scheme = "http"
+	}
+	return fmt.Sprintf("%s://%s:%d", scheme, healthCheckHostname, grafanaContainerPort)
+}
+
 func clusterUniqueName(cr *operatorv1beta1.Cryostat) string {
 	// Use the SHA256 checksum of the namespaced name as a suffix
 	nn := types.NamespacedName{Namespace: cr.Namespace, Name: cr.Name}
 	suffix := fmt.Sprintf("%x", sha256.Sum256([]byte(nn.String())))
 	return "cryostat-" + suffix
 }
+
+// Matches image tags of the form "major.minor.patch"
+var develVerRegexp = regexp.MustCompile(`(?i)(:latest|SNAPSHOT|dev|BETA\d+)$`)
+
+func getPullPolicy(imageTag string) corev1.PullPolicy {
+	// Use Always for tags that have a known development suffix
+	if develVerRegexp.MatchString(imageTag) {
+		return corev1.PullAlways
+	}
+	// Likely a release, use IfNotPresent
+	return corev1.PullIfNotPresent
+}
+
+func newVolumeForCR(cr *operatorv1beta1.Cryostat) []corev1.Volume {
+	var volumeSource corev1.VolumeSource
+	deployEmptyDir := cr.Spec.StorageOptions != nil && cr.Spec.StorageOptions.EmptyDir != nil && cr.Spec.StorageOptions.EmptyDir.Enabled
+
+	if deployEmptyDir {
+		emptyDir := cr.Spec.StorageOptions.EmptyDir
+
+		sizeLimit, err := resource.ParseQuantity(emptyDir.SizeLimit)
+		if err != nil {
+			sizeLimit = *resource.NewQuantity(0, resource.BinarySI)
+		}
+
+		volumeSource = corev1.VolumeSource{
+			EmptyDir: &corev1.EmptyDirVolumeSource{
+				Medium:    emptyDir.Medium,
+				SizeLimit: &sizeLimit,
+			},
+		}
+	} else {
+		volumeSource = corev1.VolumeSource{
+			PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
+				ClaimName: cr.Name,
+			},
+		}
+	}
+
+	return []corev1.Volume{
+		{
+			Name:         cr.Name,
+			VolumeSource: volumeSource,
+		},
+	}
+}
