@@ -57,6 +57,7 @@ import (
 
 	"github.com/cryostatio/cryostat-operator/internal/controllers/common"
 	resources "github.com/cryostatio/cryostat-operator/internal/controllers/common/resource_definitions"
+	configv1 "github.com/openshift/api/config/v1"
 	consolev1 "github.com/openshift/api/console/v1"
 	openshiftv1 "github.com/openshift/api/route/v1"
 	securityv1 "github.com/openshift/api/security/v1"
@@ -97,10 +98,38 @@ const datasourceImageTagEnv = "RELATED_IMAGE_DATASOURCE"
 // Environment variable to override the Grafana dashboard image
 const grafanaImageTagEnv = "RELATED_IMAGE_GRAFANA"
 
+// Environment variable to override the cryostat-reports image
+const reportsImageTagEnv = "RELATED_IMAGE_REPORTS"
+
 // Regular expression for the start of a GID range in the OpenShift
 // supplemental groups SCC annotation
 var supGroupRegexp = regexp.MustCompile(`^\d+`)
 
+// The canonical name of an APIServer instance
+const apiServerName = "cluster"
+
+// Reasons for Cryostat Conditions
+const (
+	reasonWaitingForCert         = "WaitingForCertificate"
+	reasonAllCertsReady          = "AllCertificatesReady"
+	reasonCertManagerUnavailable = "CertManagerUnavailable"
+	reasonCertManagerDisabled    = "CertManagerDisabled"
+)
+
+// Map Cryostat conditions to deployment conditions
+type deploymentConditionTypeMap map[operatorv1beta1.CryostatConditionType]appsv1.DeploymentConditionType
+
+var mainDeploymentConditions = deploymentConditionTypeMap{
+	operatorv1beta1.ConditionTypeMainDeploymentAvailable:      appsv1.DeploymentAvailable,
+	operatorv1beta1.ConditionTypeMainDeploymentProgressing:    appsv1.DeploymentProgressing,
+	operatorv1beta1.ConditionTypeMainDeploymentReplicaFailure: appsv1.DeploymentReplicaFailure,
+}
+var reportsDeploymentConditions = deploymentConditionTypeMap{
+	operatorv1beta1.ConditionTypeReportsDeploymentAvailable:      appsv1.DeploymentAvailable,
+	operatorv1beta1.ConditionTypeReportsDeploymentProgressing:    appsv1.DeploymentProgressing,
+	operatorv1beta1.ConditionTypeReportsDeploymentReplicaFailure: appsv1.DeploymentReplicaFailure,
+}
+
 // +kubebuilder:rbac:namespace=system,groups="",resources=pods;services;services/finalizers;endpoints;persistentvolumeclaims;events;configmaps;secrets;serviceaccounts,verbs=*
 // +kubebuilder:rbac:namespace=system,groups="",resources=replicationcontrollers,verbs=get
 // +kubebuilder:rbac:namespace=system,groups=rbac.authorization.k8s.io,resources=roles;rolebindings,verbs=create;get;list;update;watch;delete
@@ -108,6 +137,8 @@ var supGroupRegexp = regexp.MustCompile(`^\d+`)
 // +kubebuilder:rbac:groups=authentication.k8s.io,resources=tokenreviews,verbs=create
 // +kubebuilder:rbac:groups=authorization.k8s.io,resources=selfsubjectaccessreviews,verbs=create
 // +kubebuilder:rbac:groups="",resources=namespaces,verbs=get;list;watch
+// +kubebuilder:rbac:groups=oauth.openshift.io,resources=oauthaccesstokens,verbs=list;delete
+// +kubebuilder:rbac:groups=config.openshift.io,resources=apiservers,verbs=get;list;update;watch
 // +kubebuilder:rbac:namespace=system,groups=route.openshift.io,resources=routes;routes/custom-host,verbs=*
 // +kubebuilder:rbac:namespace=system,groups=apps.openshift.io,resources=deploymentconfigs,verbs=get
 // +kubebuilder:rbac:namespace=system,groups=apps,resources=deployments;daemonsets;replicasets;statefulsets,verbs=*
@@ -154,6 +185,11 @@ func (r *CryostatReconciler) Reconcile(ctx context.Context, request ctrl.Request
 				if err != nil {
 					return reconcile.Result{}, err
 				}
+
+				err = r.deleteCorsAllowedOrigins(ctx, instance.Status.ApplicationURL, instance)
+				if err != nil {
+					return reconcile.Result{}, err
+				}
 			}
 
 			err = common.RemoveFinalizer(ctx, r.Client, instance, cryostatFinalizer)
@@ -175,12 +211,15 @@ func (r *CryostatReconciler) Reconcile(ctx context.Context, request ctrl.Request
 
 	reqLogger.Info("Spec", "Minimal", instance.Spec.Minimal)
 
-	pvc := resources.NewPersistentVolumeClaimForCR(instance)
-	if err := controllerutil.SetControllerReference(instance, pvc, r.Scheme); err != nil {
-		return reconcile.Result{}, err
-	}
-	if err = r.createObjectIfNotExists(context.Background(), types.NamespacedName{Name: pvc.Name, Namespace: pvc.Namespace}, &corev1.PersistentVolumeClaim{}, pvc); err != nil {
-		return reconcile.Result{}, err
+	shouldCreatePvc := !(instance.Spec.StorageOptions != nil && instance.Spec.StorageOptions.EmptyDir != nil && instance.Spec.StorageOptions.EmptyDir.Enabled)
+	if shouldCreatePvc {
+		pvc := resources.NewPersistentVolumeClaimForCR(instance)
+		if err := controllerutil.SetControllerReference(instance, pvc, r.Scheme); err != nil {
+			return reconcile.Result{}, err
+		}
+		if err = r.createObjectIfNotExists(context.Background(), types.NamespacedName{Name: pvc.Name, Namespace: pvc.Namespace}, &corev1.PersistentVolumeClaim{}, pvc); err != nil {
+			return reconcile.Result{}, err
+		}
 	}
 
 	grafanaSecret := resources.NewGrafanaSecretForCR(instance)
@@ -206,8 +245,17 @@ func (r *CryostatReconciler) Reconcile(ctx context.Context, request ctrl.Request
 		tlsConfig, err = r.setupTLS(context.Background(), instance)
 		if err != nil {
 			if err == common.ErrCertNotReady {
+				condErr := r.updateCondition(ctx, instance, operatorv1beta1.ConditionTypeTLSSetupComplete, metav1.ConditionFalse,
+					reasonWaitingForCert, "Waiting for certificates to become ready.")
+				if condErr != nil {
+					return reconcile.Result{}, err
+				}
 				return reconcile.Result{RequeueAfter: 5 * time.Second}, nil
 			}
+			if err == errCertManagerMissing {
+				r.updateCondition(ctx, instance, operatorv1beta1.ConditionTypeTLSSetupComplete, metav1.ConditionFalse,
+					reasonCertManagerUnavailable, eventCertManagerUnavailableMsg)
+			}
 			reqLogger.Error(err, "Failed to set up TLS for Cryostat")
 			return reconcile.Result{}, err
 		}
@@ -221,6 +269,18 @@ func (r *CryostatReconciler) Reconcile(ctx context.Context, request ctrl.Request
 			Termination:              openshiftv1.TLSTerminationReencrypt,
 			DestinationCACertificate: string(caCert),
 		}
+
+		err = r.updateCondition(ctx, instance, operatorv1beta1.ConditionTypeTLSSetupComplete, metav1.ConditionTrue,
+			reasonAllCertsReady, "All certificates for Cryostat components are ready.")
+		if err != nil {
+			return reconcile.Result{}, err
+		}
+	} else {
+		err = r.updateCondition(ctx, instance, operatorv1beta1.ConditionTypeTLSSetupComplete, metav1.ConditionTrue,
+			reasonCertManagerDisabled, "TLS setup has been disabled.")
+		if err != nil {
+			return reconcile.Result{}, err
+		}
 	}
 
 	// Create RBAC resources for Cryostat
@@ -232,11 +292,11 @@ func (r *CryostatReconciler) Reconcile(ctx context.Context, request ctrl.Request
 	serviceSpecs := &resources.ServiceSpecs{}
 	if !instance.Spec.Minimal {
 		grafanaSvc := resources.NewGrafanaService(instance)
-		url, err := r.createService(context.Background(), instance, grafanaSvc, &grafanaSvc.Spec.Ports[0], routeTLS)
+		svcUrl, err := r.createService(context.Background(), instance, grafanaSvc, &grafanaSvc.Spec.Ports[0], routeTLS)
 		if err != nil {
 			return requeueIfIngressNotReady(reqLogger, err)
 		}
-		serviceSpecs.GrafanaURL = url
+		serviceSpecs.GrafanaURL = svcUrl
 
 		// check for existing minimal deployment and delete if found
 		deployment := &appsv1.Deployment{}
@@ -303,25 +363,24 @@ func (r *CryostatReconciler) Reconcile(ctx context.Context, request ctrl.Request
 		}
 	}
 
-	exporterSvc := resources.NewExporterService(instance)
-	url, err := r.createService(context.Background(), instance, exporterSvc, &exporterSvc.Spec.Ports[0], routeTLS)
+	coreSvc := resources.NewCoreService(instance)
+	svcUrl, err := r.createService(context.Background(), instance, coreSvc, &coreSvc.Spec.Ports[0], routeTLS)
 	if err != nil {
 		return requeueIfIngressNotReady(reqLogger, err)
 	}
-	serviceSpecs.CoreURL = url
-
-	cmdChanSvc := resources.NewCommandChannelService(instance)
-	url, err = r.createService(context.Background(), instance, cmdChanSvc, &cmdChanSvc.Spec.Ports[0], routeTLS)
-	if err != nil {
-		return requeueIfIngressNotReady(reqLogger, err)
-	}
-	serviceSpecs.CommandURL = url
+	serviceSpecs.CoreURL = svcUrl
 
 	imageTags := r.getImageTags()
 	fsGroup, err := r.getFSGroup(ctx, instance.Namespace)
 	if err != nil {
 		return reconcile.Result{}, err
 	}
+
+	reportsResult, err := r.reconcileReports(ctx, reqLogger, instance, tlsConfig, imageTags, serviceSpecs)
+	if err != nil {
+		return reportsResult, err
+	}
+
 	deployment := resources.NewDeploymentForCR(instance, serviceSpecs, imageTags, tlsConfig, *fsGroup, r.IsOpenShift)
 	podTemplate := deployment.Spec.Template.DeepCopy()
 	if err := controllerutil.SetControllerReference(instance, deployment, r.Scheme); err != nil {
@@ -351,6 +410,20 @@ func (r *CryostatReconciler) Reconcile(ctx context.Context, request ctrl.Request
 		if err != nil {
 			return reconcile.Result{}, err
 		}
+
+		if instance.Status.ApplicationURL != "" {
+			err = r.addCorsAllowedOriginIfNotPresent(ctx, instance.Status.ApplicationURL, instance)
+			if err != nil {
+				return reconcile.Result{}, err
+			}
+		}
+	}
+
+	// Check deployment status and update conditions
+	err = r.updateConditionsFromDeployment(ctx, instance, types.NamespacedName{Name: instance.Name, Namespace: instance.Namespace},
+		mainDeploymentConditions)
+	if err != nil {
+		return reconcile.Result{}, err
 	}
 
 	reqLogger.Info("Successfully reconciled deployment", "Deployment.Namespace", deployment.Namespace, "Deployment.Name", deployment.Name)
@@ -379,6 +452,78 @@ func (r *CryostatReconciler) SetupWithManager(mgr ctrl.Manager) error {
 	return c.Complete(r)
 }
 
+func (r *CryostatReconciler) reconcileReports(ctx context.Context, reqLogger logr.Logger, instance *operatorv1beta1.Cryostat,
+	tls *resources.TLSConfig, imageTags *resources.ImageTags, serviceSpecs *resources.ServiceSpecs) (reconcile.Result, error) {
+	reqLogger.Info("Spec", "Reports", instance.Spec.ReportOptions)
+
+	if instance.Spec.ReportOptions == nil {
+		instance.Spec.ReportOptions = &operatorv1beta1.ReportConfiguration{Replicas: 0}
+	}
+	desired := instance.Spec.ReportOptions.Replicas
+
+	deployment := resources.NewDeploymentForReports(instance, imageTags, tls)
+	if desired == 0 {
+		svc := resources.NewReportService(instance)
+		if err := r.Client.Delete(ctx, svc); err != nil && !errors.IsNotFound(err) {
+			return reconcile.Result{}, err
+		}
+		if err := r.Client.Delete(ctx, deployment); err != nil && !errors.IsNotFound(err) {
+			return reconcile.Result{}, err
+		}
+
+		removeConditionIfPresent(instance, operatorv1beta1.ConditionTypeReportsDeploymentAvailable,
+			operatorv1beta1.ConditionTypeReportsDeploymentProgressing,
+			operatorv1beta1.ConditionTypeReportsDeploymentReplicaFailure)
+		err := r.Client.Status().Update(ctx, instance)
+		if err != nil {
+			return reconcile.Result{}, err
+		}
+		return reconcile.Result{}, nil
+	}
+
+	if desired > 0 {
+		svc := resources.NewReportService(instance)
+		if err := controllerutil.SetControllerReference(instance, svc, r.Scheme); err != nil {
+			return reconcile.Result{}, err
+		}
+		if err := r.createObjectIfNotExists(ctx, types.NamespacedName{Name: svc.Name, Namespace: svc.Namespace}, &corev1.Service{}, svc); err != nil {
+			return reconcile.Result{}, err
+		}
+
+		if err := controllerutil.SetControllerReference(instance, deployment, r.Scheme); err != nil {
+			return reconcile.Result{}, err
+		}
+
+		podTemplate := deployment.Spec.Template.DeepCopy()
+		op, err := controllerutil.CreateOrUpdate(ctx, r.Client, deployment, func() error {
+			deployment.Spec.Template.Spec = podTemplate.Spec
+			deployment.Spec.Replicas = &desired
+			return nil
+		})
+		if err != nil {
+			return reconcile.Result{}, err
+		}
+
+		scheme := "https"
+		if tls == nil {
+			scheme = "http"
+		}
+		serviceSpecs.ReportsURL = &url.URL{
+			Scheme: scheme,
+			Host:   svc.Name + ":" + strconv.Itoa(int(svc.Spec.Ports[0].Port)),
+		}
+		reqLogger.Info(fmt.Sprintf("Reports Deployment %s", op))
+
+		// Check deployment status and update conditions
+		err = r.updateConditionsFromDeployment(ctx, instance, types.NamespacedName{Name: deployment.Name, Namespace: deployment.Namespace},
+			reportsDeploymentConditions)
+		if err != nil {
+			return reconcile.Result{}, err
+		}
+	}
+	return reconcile.Result{}, nil
+}
+
 func (r *CryostatReconciler) createService(ctx context.Context, controller *operatorv1beta1.Cryostat, svc *corev1.Service, exposePort *corev1.ServicePort,
 	tlsConfig *openshiftv1.TLSConfig) (*url.URL, error) {
 	if err := controllerutil.SetControllerReference(controller, svc, r.Scheme); err != nil {
@@ -526,14 +671,36 @@ func (r *CryostatReconciler) createObjectIfNotExists(ctx context.Context, ns typ
 
 func (r *CryostatReconciler) createRBAC(ctx context.Context, cr *operatorv1beta1.Cryostat) error {
 	// Create ServiceAccount
-	sa := resources.NewServiceAccountForCR(cr)
-	if err := controllerutil.SetControllerReference(cr, sa, r.Scheme); err != nil {
+	sa, err := resources.NewServiceAccountForCR(cr, r.IsOpenShift)
+	if err != nil {
 		return err
 	}
-	if err := r.createObjectIfNotExists(ctx, types.NamespacedName{Name: sa.Name, Namespace: sa.Namespace},
-		&corev1.ServiceAccount{}, sa); err != nil {
+	newSA := sa.DeepCopy()
+	op, err := controllerutil.CreateOrUpdate(ctx, r.Client, sa, func() error {
+		if err := controllerutil.SetControllerReference(cr, sa, r.Scheme); err != nil {
+			return err
+		}
+		// TODO just replace the labels and annotations we manage, once we allow the user to configure
+		// ServiceAccount annotations/labels in the CR, we can simply overwrite them all
+		for key, val := range newSA.GetAnnotations() {
+			metav1.SetMetaDataAnnotation(&sa.ObjectMeta, key, val)
+		}
+		if sa.Labels == nil {
+			sa.Labels = map[string]string{}
+		}
+		for key, val := range newSA.GetLabels() {
+			// TODO use metav1.SetMetaDataLabel when updating client-go, replace above initialization
+			sa.Labels[key] = val
+		}
+		// Pod needs SA token, do not allow to be disabled
+		sa.AutomountServiceAccountToken = newSA.AutomountServiceAccountToken
+		// Secrets, ImagePullSecrets are modified by Kubernetes/OpenShift
+		return nil
+	})
+	if err != nil {
 		return err
 	}
+	r.Log.Info(fmt.Sprintf("ServiceAccount %s", op), "name", sa.Name, "namespace", sa.Namespace)
 
 	// Create Role
 	role := resources.NewRoleForCR(cr)
@@ -588,6 +755,7 @@ func (r *CryostatReconciler) getImageTags() *resources.ImageTags {
 		CoreImageTag:       r.getEnvOrDefault(coreImageTagEnv, resources.DefaultCoreImageTag),
 		DatasourceImageTag: r.getEnvOrDefault(datasourceImageTagEnv, resources.DefaultDatasourceImageTag),
 		GrafanaImageTag:    r.getEnvOrDefault(grafanaImageTagEnv, resources.DefaultGrafanaImageTag),
+		ReportsImageTag:    r.getEnvOrDefault(reportsImageTagEnv, resources.DefaultReportsImageTag),
 	}
 }
 
@@ -655,6 +823,114 @@ func (r *CryostatReconciler) deleteConsoleLink(ctx context.Context, cr *operator
 	return nil
 }
 
+func (r *CryostatReconciler) addCorsAllowedOriginIfNotPresent(ctx context.Context, allowedOrigin string, cr *operatorv1beta1.Cryostat) error {
+	reqLogger := r.Log.WithValues("Request.Namespace", cr.Namespace, "Request.Name", cr.Name)
+	apiServer := &configv1.APIServer{}
+	err := r.Client.Get(context.Background(), types.NamespacedName{Name: apiServerName}, apiServer)
+	if err != nil {
+		reqLogger.Error(err, "Failed to get APIServer config")
+		return err
+	}
+
+	allowedOriginAsRegex := regexp.QuoteMeta(allowedOrigin)
+
+	for _, origin := range apiServer.Spec.AdditionalCORSAllowedOrigins {
+		if origin == allowedOriginAsRegex {
+			return nil
+		}
+	}
+
+	apiServer.Spec.AdditionalCORSAllowedOrigins = append(
+		apiServer.Spec.AdditionalCORSAllowedOrigins,
+		allowedOriginAsRegex,
+	)
+
+	err = r.Client.Update(ctx, apiServer)
+	if err != nil {
+		reqLogger.Error(err, "Failed to update APIServer CORS allowed origins")
+		return err
+	}
+
+	return nil
+}
+
+func (r *CryostatReconciler) deleteCorsAllowedOrigins(ctx context.Context, allowedOrigin string, cr *operatorv1beta1.Cryostat) error {
+	reqLogger := r.Log.WithValues("Request.Namespace", cr.Namespace, "Request.Name", cr.Name)
+	apiServer := &configv1.APIServer{}
+	err := r.Client.Get(context.Background(), types.NamespacedName{Name: apiServerName}, apiServer)
+	if err != nil {
+		reqLogger.Error(err, "Failed to get APIServer config")
+		return err
+	}
+
+	allowedOriginAsRegex := regexp.QuoteMeta(allowedOrigin)
+
+	for i, origin := range apiServer.Spec.AdditionalCORSAllowedOrigins {
+		if origin == allowedOriginAsRegex {
+			apiServer.Spec.AdditionalCORSAllowedOrigins = append(
+				apiServer.Spec.AdditionalCORSAllowedOrigins[:i],
+				apiServer.Spec.AdditionalCORSAllowedOrigins[i+1:]...)
+			break
+		}
+	}
+
+	err = r.Client.Update(ctx, apiServer)
+	if err != nil {
+		reqLogger.Error(err, "Failed to remove Cryostat origin from APIServer CORS allowed origins")
+		return err
+	}
+
+	return nil
+}
+
+func (r *CryostatReconciler) updateCondition(ctx context.Context, cr *operatorv1beta1.Cryostat,
+	condType operatorv1beta1.CryostatConditionType, status metav1.ConditionStatus, reason string, message string) error {
+	reqLogger := r.Log.WithValues("Request.Namespace", cr.Namespace, "Request.Name", cr.Name)
+	meta.SetStatusCondition(&cr.Status.Conditions, metav1.Condition{
+		Type:    string(condType),
+		Status:  status,
+		Reason:  reason,
+		Message: message,
+	})
+	err := r.Client.Status().Update(ctx, cr)
+	if err != nil {
+		reqLogger.Error(err, "failed to update condition", "type", condType)
+	}
+	return err
+}
+
+func (r *CryostatReconciler) updateConditionsFromDeployment(ctx context.Context, cr *operatorv1beta1.Cryostat,
+	deployKey types.NamespacedName, mapping deploymentConditionTypeMap) error {
+	reqLogger := r.Log.WithValues("Request.Namespace", cr.Namespace, "Request.Name", cr.Name)
+
+	// Get deployment's latest conditions
+	deploy := &appsv1.Deployment{}
+	err := r.Client.Get(ctx, types.NamespacedName{Name: deployKey.Name, Namespace: deployKey.Namespace}, deploy)
+	if err != nil {
+		return err
+	}
+
+	// Associate deployment conditions with Cryostat conditions
+	for condType, deployCondType := range mapping {
+		condition := findDeployCondition(deploy.Status.Conditions, deployCondType)
+		if condition == nil {
+			removeConditionIfPresent(cr, condType)
+		} else {
+			meta.SetStatusCondition(&cr.Status.Conditions, metav1.Condition{
+				Type:    string(condType),
+				Status:  metav1.ConditionStatus(condition.Status),
+				Reason:  condition.Reason,
+				Message: condition.Message,
+			})
+		}
+	}
+	err = r.Client.Status().Update(ctx, cr)
+	if err != nil {
+		reqLogger.Error(err, "failed to update conditions for deployment", "deployment", deploy.Name)
+	}
+	return err
+}
+
 func getProtocol(tlsConfig *openshiftv1.TLSConfig) string {
 	if tlsConfig == nil {
 		return "http"
@@ -673,11 +949,27 @@ func requeueIfIngressNotReady(log logr.Logger, err error) (reconcile.Result, err
 func getNetworkConfig(controller *operatorv1beta1.Cryostat, svc *corev1.Service) (*operatorv1beta1.NetworkConfiguration, error) {
 	if svc.Name == controller.Name {
 		return controller.Spec.NetworkOptions.CoreConfig, nil
-	} else if svc.Name == controller.Name+"-command" {
-		return controller.Spec.NetworkOptions.CommandConfig, nil
 	} else if svc.Name == controller.Name+"-grafana" {
 		return controller.Spec.NetworkOptions.GrafanaConfig, nil
 	} else {
 		return nil, goerrors.New("Service name not recognized")
 	}
 }
+
+func removeConditionIfPresent(cr *operatorv1beta1.Cryostat, condType ...operatorv1beta1.CryostatConditionType) {
+	for _, ct := range condType {
+		found := meta.FindStatusCondition(cr.Status.Conditions, string(ct))
+		if found != nil {
+			meta.RemoveStatusCondition(&cr.Status.Conditions, string(ct))
+		}
+	}
+}
+
+func findDeployCondition(conditions []appsv1.DeploymentCondition, condType appsv1.DeploymentConditionType) *appsv1.DeploymentCondition {
+	for _, condition := range conditions {
+		if condition.Type == condType {
+			return &condition
+		}
+	}
+	return nil
+}
