@@ -42,7 +42,9 @@ import (
 	operatorv1beta1 "github.com/cryostatio/cryostat-operator/api/v1beta1"
 	"github.com/cryostatio/cryostat-operator/internal/controllers/common/resource_definitions"
 	certv1 "github.com/jetstack/cert-manager/pkg/apis/certmanager/v1"
+	certMeta "github.com/jetstack/cert-manager/pkg/apis/meta/v1"
 	"github.com/onsi/gomega"
+	configv1 "github.com/openshift/api/config/v1"
 	consolev1 "github.com/openshift/api/console/v1"
 	routev1 "github.com/openshift/api/route/v1"
 	securityv1 "github.com/openshift/api/security/v1"
@@ -134,7 +136,14 @@ func NewCryostatWithTemplates() *operatorv1beta1.Cryostat {
 
 func NewCryostatWithIngress() *operatorv1beta1.Cryostat {
 	cr := NewCryostat()
-	networkConfig := NewNetworkConfigurationList()
+	networkConfig := NewNetworkConfigurationList(true)
+	cr.Spec.NetworkOptions = &networkConfig
+	return cr
+}
+
+func NewCryostatWithIngressNoTLS() *operatorv1beta1.Cryostat {
+	cr := NewCryostat()
+	networkConfig := NewNetworkConfigurationList(false)
 	cr.Spec.NetworkOptions = &networkConfig
 	return cr
 }
@@ -178,6 +187,116 @@ func NewCryostatWithPVCLabelsOnly() *operatorv1beta1.Cryostat {
 	return cr
 }
 
+func NewCryostatWithDefaultEmptyDir() *operatorv1beta1.Cryostat {
+	cr := NewCryostat()
+	cr.Spec.StorageOptions = &operatorv1beta1.StorageConfiguration{
+		EmptyDir: &operatorv1beta1.EmptyDirConfig{
+			Enabled: true,
+		},
+	}
+	return cr
+}
+
+func NewCryostatWithEmptyDirSpec() *operatorv1beta1.Cryostat {
+	cr := NewCryostat()
+	cr.Spec.StorageOptions = &operatorv1beta1.StorageConfiguration{
+		EmptyDir: &operatorv1beta1.EmptyDirConfig{
+			Enabled:   true,
+			Medium:    "Memory",
+			SizeLimit: "200Mi",
+		},
+	}
+	return cr
+}
+
+func NewCryostatWithCoreSvc() *operatorv1beta1.Cryostat {
+	svcType := corev1.ServiceTypeNodePort
+	httpPort := int32(8080)
+	jmxPort := int32(9095)
+	cr := NewCryostat()
+	cr.Spec.ServiceOptions = &operatorv1beta1.ServiceConfigList{
+		CoreConfig: &operatorv1beta1.CoreServiceConfig{
+			HTTPPort: &httpPort,
+			JMXPort:  &jmxPort,
+			ServiceConfig: operatorv1beta1.ServiceConfig{
+				ServiceType: &svcType,
+				Annotations: map[string]string{
+					"my/custom": "annotation",
+				},
+				Labels: map[string]string{
+					"my":  "label",
+					"app": "somethingelse",
+				},
+			},
+		},
+	}
+	return cr
+}
+
+func NewCryostatWithGrafanaSvc() *operatorv1beta1.Cryostat {
+	svcType := corev1.ServiceTypeNodePort
+	httpPort := int32(8080)
+	cr := NewCryostat()
+	cr.Spec.ServiceOptions = &operatorv1beta1.ServiceConfigList{
+		GrafanaConfig: &operatorv1beta1.GrafanaServiceConfig{
+			HTTPPort: &httpPort,
+			ServiceConfig: operatorv1beta1.ServiceConfig{
+				ServiceType: &svcType,
+				Annotations: map[string]string{
+					"my/custom": "annotation",
+				},
+				Labels: map[string]string{
+					"my":  "label",
+					"app": "somethingelse",
+				},
+			},
+		},
+	}
+	return cr
+}
+
+func NewCryostatWithReportsSvc() *operatorv1beta1.Cryostat {
+	svcType := corev1.ServiceTypeNodePort
+	httpPort := int32(13161)
+	cr := NewCryostat()
+	cr.Spec.ReportOptions = &operatorv1beta1.ReportConfiguration{
+		Replicas: 1,
+	}
+	cr.Spec.ServiceOptions = &operatorv1beta1.ServiceConfigList{
+		ReportsConfig: &operatorv1beta1.ReportsServiceConfig{
+			HTTPPort: &httpPort,
+			ServiceConfig: operatorv1beta1.ServiceConfig{
+				ServiceType: &svcType,
+				Annotations: map[string]string{
+					"my/custom": "annotation",
+				},
+				Labels: map[string]string{
+					"my":  "label",
+					"app": "somethingelse",
+				},
+			},
+		},
+	}
+	return cr
+}
+
+func NewCryostatWithReportsResources() *operatorv1beta1.Cryostat {
+	cr := NewCryostat()
+	cr.Spec.ReportOptions = &operatorv1beta1.ReportConfiguration{
+		Resources: corev1.ResourceRequirements{
+			Limits: corev1.ResourceList{
+				corev1.ResourceCPU:    resource.MustParse("1600m"),
+				corev1.ResourceMemory: resource.MustParse("1Gi"),
+			},
+			Requests: corev1.ResourceList{
+				corev1.ResourceCPU:    resource.MustParse("800m"),
+				corev1.ResourceMemory: resource.MustParse("512Mi"),
+			},
+		},
+	}
+	return cr
+}
+
 func NewCryostatCertManagerDisabled() *operatorv1beta1.Cryostat {
 	cr := NewCryostat()
 	certManager := false
@@ -191,6 +310,43 @@ func NewCryostatCertManagerUndefined() *operatorv1beta1.Cryostat {
 	return cr
 }
 
+func NewCryostatWithResources() *operatorv1beta1.Cryostat {
+	cr := NewCryostat()
+	cr.Spec.Resources = operatorv1beta1.ResourceConfigList{
+		CoreResources: corev1.ResourceRequirements{
+			Limits: corev1.ResourceList{
+				corev1.ResourceCPU:    resource.MustParse("500m"),
+				corev1.ResourceMemory: resource.MustParse("256Mi"),
+			},
+			Requests: corev1.ResourceList{
+				corev1.ResourceCPU:    resource.MustParse("250m"),
+				corev1.ResourceMemory: resource.MustParse("128Mi"),
+			},
+		},
+		GrafanaResources: corev1.ResourceRequirements{
+			Limits: corev1.ResourceList{
+				corev1.ResourceCPU:    resource.MustParse("550m"),
+				corev1.ResourceMemory: resource.MustParse("512Mi"),
+			},
+			Requests: corev1.ResourceList{
+				corev1.ResourceCPU:    resource.MustParse("128m"),
+				corev1.ResourceMemory: resource.MustParse("256Mi"),
+			},
+		},
+		DataSourceResources: corev1.ResourceRequirements{
+			Limits: corev1.ResourceList{
+				corev1.ResourceCPU:    resource.MustParse("600m"),
+				corev1.ResourceMemory: resource.MustParse("128Mi"),
+			},
+			Requests: corev1.ResourceList{
+				corev1.ResourceCPU:    resource.MustParse("300m"),
+				corev1.ResourceMemory: resource.MustParse("64Mi"),
+			},
+		},
+	}
+	return cr
+}
+
 func newPVCSpec(storageClass string, storageRequest string,
 	accessModes ...corev1.PersistentVolumeAccessMode) *corev1.PersistentVolumeClaimSpec {
 	return &corev1.PersistentVolumeClaimSpec{
@@ -210,6 +366,29 @@ func NewMinimalCryostat() *operatorv1beta1.Cryostat {
 	return cr
 }
 
+func NewCryostatWithJmxCacheOptionsSpec() *operatorv1beta1.Cryostat {
+	cr := NewCryostat()
+	cr.Spec.JmxCacheOptions = &operatorv1beta1.JmxCacheOptions{
+		TargetCacheSize: 10,
+		TargetCacheTTL:  20,
+	}
+	return cr
+}
+
+func NewCryostatWithWsConnectionsSpec() *operatorv1beta1.Cryostat {
+	cr := NewCryostat()
+	cr.Spec.MaxWsConnections = 10
+	return cr
+}
+
+func NewCryostatWithReportSubprocessHeapSpec() *operatorv1beta1.Cryostat {
+	cr := NewCryostat()
+	cr.Spec.ReportOptions = &operatorv1beta1.ReportConfiguration{
+		SubProcessMaxHeapSize: 500,
+	}
+	return cr
+}
+
 func NewFlightRecorder() *operatorv1beta1.FlightRecorder {
 	return newFlightRecorder(&operatorv1beta1.JMXAuthSecret{
 		SecretName: "test-jmx-auth",
@@ -534,19 +713,153 @@ func NewCryostatService() *corev1.Service {
 					Controller: &c,
 				},
 			},
+			Labels: map[string]string{
+				"app":       "cryostat",
+				"component": "cryostat",
+			},
 		},
 		Spec: corev1.ServiceSpec{
-			ClusterIP: "1.2.3.4",
+			Type: corev1.ServiceTypeClusterIP,
+			Selector: map[string]string{
+				"app":       "cryostat",
+				"component": "cryostat",
+			},
 			Ports: []corev1.ServicePort{
 				{
-					Name: "export",
-					Port: 8181,
+					Name:       "http",
+					Port:       8181,
+					TargetPort: intstr.FromInt(8181),
+				},
+				{
+					Name:       "jfr-jmx",
+					Port:       9091,
+					TargetPort: intstr.FromInt(9091),
 				},
 			},
 		},
 	}
 }
 
+func NewGrafanaService() *corev1.Service {
+	c := true
+	return &corev1.Service{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "cryostat-grafana",
+			Namespace: "default",
+			OwnerReferences: []metav1.OwnerReference{
+				{
+					APIVersion: operatorv1beta1.GroupVersion.String(),
+					Kind:       "Cryostat",
+					Name:       "cryostat",
+					UID:        "",
+					Controller: &c,
+				},
+			},
+			Labels: map[string]string{
+				"app":       "cryostat",
+				"component": "cryostat",
+			},
+		},
+		Spec: corev1.ServiceSpec{
+			Type: corev1.ServiceTypeClusterIP,
+			Selector: map[string]string{
+				"app":       "cryostat",
+				"component": "cryostat",
+			},
+			Ports: []corev1.ServicePort{
+				{
+					Name:       "http",
+					Port:       3000,
+					TargetPort: intstr.FromInt(3000),
+				},
+			},
+		},
+	}
+}
+
+func NewReportsService() *corev1.Service {
+	c := true
+	return &corev1.Service{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "cryostat-reports",
+			Namespace: "default",
+			OwnerReferences: []metav1.OwnerReference{
+				{
+					APIVersion: operatorv1beta1.GroupVersion.String(),
+					Kind:       "Cryostat",
+					Name:       "cryostat-reports",
+					UID:        "",
+					Controller: &c,
+				},
+			},
+			Labels: map[string]string{
+				"app":       "cryostat",
+				"component": "reports",
+			},
+		},
+		Spec: corev1.ServiceSpec{
+			Type: corev1.ServiceTypeClusterIP,
+			Selector: map[string]string{
+				"app":       "cryostat",
+				"component": "reports",
+			},
+			Ports: []corev1.ServicePort{
+				{
+					Name:       "http",
+					Port:       10000,
+					TargetPort: intstr.FromInt(10000),
+				},
+			},
+		},
+	}
+}
+
+func NewCustomizedCoreService() *corev1.Service {
+	svc := NewCryostatService()
+	svc.Spec.Type = corev1.ServiceTypeNodePort
+	svc.Spec.Ports[0].Port = 8080
+	svc.Spec.Ports[1].Port = 9095
+	svc.Annotations = map[string]string{
+		"my/custom": "annotation",
+	}
+	svc.Labels = map[string]string{
+		"app":       "cryostat",
+		"component": "cryostat",
+		"my":        "label",
+	}
+	return svc
+}
+
+func NewCustomizedGrafanaService() *corev1.Service {
+	svc := NewGrafanaService()
+	svc.Spec.Type = corev1.ServiceTypeNodePort
+	svc.Spec.Ports[0].Port = 8080
+	svc.Annotations = map[string]string{
+		"my/custom": "annotation",
+	}
+	svc.Labels = map[string]string{
+		"app":       "cryostat",
+		"component": "cryostat",
+		"my":        "label",
+	}
+	return svc
+}
+
+func NewCustomizedReportsService() *corev1.Service {
+	svc := NewReportsService()
+	svc.Spec.Type = corev1.ServiceTypeNodePort
+	svc.Spec.Ports[0].Port = 13161
+	svc.Annotations = map[string]string{
+		"my/custom": "annotation",
+	}
+	svc.Labels = map[string]string{
+		"app":       "cryostat",
+		"component": "reports",
+		"my":        "label",
+	}
+	return svc
+}
+
 func NewTestService() *corev1.Service {
 	return &corev1.Service{
 		ObjectMeta: metav1.ObjectMeta{
@@ -565,6 +878,97 @@ func NewTestService() *corev1.Service {
 	}
 }
 
+func NewCryostatCert() *certv1.Certificate {
+	return &certv1.Certificate{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "cryostat",
+			Namespace: "default",
+		},
+		Spec: certv1.CertificateSpec{
+			CommonName: "cryostat.default.svc",
+			DNSNames: []string{
+				"cryostat",
+				"cryostat.default.svc",
+				"cryostat.default.svc.cluster.local",
+			},
+			SecretName: "cryostat-tls",
+			Keystores: &certv1.CertificateKeystores{
+				PKCS12: &certv1.PKCS12Keystore{
+					Create: true,
+					PasswordSecretRef: certMeta.SecretKeySelector{
+						LocalObjectReference: certMeta.LocalObjectReference{
+							Name: "cryostat-keystore",
+						},
+						Key: "KEYSTORE_PASS",
+					},
+				},
+			},
+			IssuerRef: certMeta.ObjectReference{
+				Name: "cryostat-ca",
+			},
+			Usages: []certv1.KeyUsage{
+				certv1.UsageDigitalSignature,
+				certv1.UsageKeyEncipherment,
+				certv1.UsageServerAuth,
+				certv1.UsageClientAuth,
+			},
+		},
+	}
+}
+
+func NewGrafanaCert() *certv1.Certificate {
+	return &certv1.Certificate{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "cryostat-grafana",
+			Namespace: "default",
+		},
+		Spec: certv1.CertificateSpec{
+			CommonName: "cryostat-grafana.default.svc",
+			DNSNames: []string{
+				"cryostat-grafana",
+				"cryostat-grafana.default.svc",
+				"cryostat-grafana.default.svc.cluster.local",
+				"cryostat-health.local",
+			},
+			SecretName: "cryostat-grafana-tls",
+			IssuerRef: certMeta.ObjectReference{
+				Name: "cryostat-ca",
+			},
+			Usages: []certv1.KeyUsage{
+				certv1.UsageDigitalSignature,
+				certv1.UsageKeyEncipherment,
+				certv1.UsageServerAuth,
+			},
+		},
+	}
+}
+
+func NewReportsCert() *certv1.Certificate {
+	return &certv1.Certificate{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "cryostat-reports",
+			Namespace: "default",
+		},
+		Spec: certv1.CertificateSpec{
+			CommonName: "cryostat-reports.default.svc",
+			DNSNames: []string{
+				"cryostat-reports",
+				"cryostat-reports.default.svc",
+				"cryostat-reports.default.svc.cluster.local",
+			},
+			SecretName: "cryostat-reports-tls",
+			IssuerRef: certMeta.ObjectReference{
+				Name: "cryostat-ca",
+			},
+			Usages: []certv1.KeyUsage{
+				certv1.UsageDigitalSignature,
+				certv1.UsageKeyEncipherment,
+				certv1.UsageServerAuth,
+			},
+		},
+	}
+}
+
 func NewCACert() *certv1.Certificate {
 	return &certv1.Certificate{
 		ObjectMeta: metav1.ObjectMeta{
@@ -572,7 +976,12 @@ func NewCACert() *certv1.Certificate {
 			Namespace: "default",
 		},
 		Spec: certv1.CertificateSpec{
+			CommonName: "ca.cryostat.cert-manager",
 			SecretName: "cryostat-ca",
+			IssuerRef: certMeta.ObjectReference{
+				Name: "cryostat-self-signed",
+			},
+			IsCA: true,
 		},
 	}
 }
@@ -589,6 +998,36 @@ func newCASecret(certData []byte) *corev1.Secret {
 	}
 }
 
+func NewSelfSignedIssuer() *certv1.Issuer {
+	return &certv1.Issuer{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "cryostat-self-signed",
+			Namespace: "default",
+		},
+		Spec: certv1.IssuerSpec{
+			IssuerConfig: certv1.IssuerConfig{
+				SelfSigned: &certv1.SelfSignedIssuer{},
+			},
+		},
+	}
+}
+
+func NewCryostatCAIssuer() *certv1.Issuer {
+	return &certv1.Issuer{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "cryostat-ca",
+			Namespace: "default",
+		},
+		Spec: certv1.IssuerSpec{
+			IssuerConfig: certv1.IssuerConfig{
+				CA: &certv1.CAIssuer{
+					SecretName: "cryostat-ca",
+				},
+			},
+		},
+	}
+}
+
 func NewJMXAuthSecret() *corev1.Secret {
 	return &corev1.Secret{
 		ObjectMeta: metav1.ObjectMeta{
@@ -688,14 +1127,26 @@ func NewDefaultPVCWithLabel() *corev1.PersistentVolumeClaim {
 	}, nil)
 }
 
+func NewDefaultEmptyDir() *corev1.EmptyDirVolumeSource {
+	sizeLimit := resource.MustParse("0")
+	return &corev1.EmptyDirVolumeSource{
+		SizeLimit: &sizeLimit,
+	}
+}
+
+func NewEmptyDirWithSpec() *corev1.EmptyDirVolumeSource {
+	sizeLimit := resource.MustParse("200Mi")
+	return &corev1.EmptyDirVolumeSource{
+		Medium:    "Memory",
+		SizeLimit: &sizeLimit,
+	}
+}
+
 func NewCorePorts() []corev1.ContainerPort {
 	return []corev1.ContainerPort{
 		{
 			ContainerPort: 8181,
 		},
-		{
-			ContainerPort: 9090,
-		},
 		{
 			ContainerPort: 9091,
 		},
@@ -718,40 +1169,24 @@ func NewDatasourcePorts() []corev1.ContainerPort {
 	}
 }
 
-func NewCoreEnvironmentVariables(minimal bool, tls bool, openshift bool) []corev1.EnvVar {
+func NewReportsPorts() []corev1.ContainerPort {
+	return []corev1.ContainerPort{
+		{
+			ContainerPort: 10000,
+		},
+	}
+}
+
+func NewCoreEnvironmentVariables(minimal bool, tls bool, externalTLS bool, openshift bool, reportsUrl string) []corev1.EnvVar {
 	envs := []corev1.EnvVar{
-		{
-			Name:  "CRYOSTAT_SSL_PROXIED",
-			Value: "true",
-		},
-		{
-			Name:  "CRYOSTAT_ALLOW_UNTRUSTED_SSL",
-			Value: "true",
-		},
 		{
 			Name:  "CRYOSTAT_WEB_PORT",
 			Value: "8181",
 		},
-		{
-			Name:  "CRYOSTAT_EXT_WEB_PORT",
-			Value: "443",
-		},
 		{
 			Name:  "CRYOSTAT_WEB_HOST",
 			Value: "cryostat.example.com",
 		},
-		{
-			Name:  "CRYOSTAT_LISTEN_PORT",
-			Value: "9090",
-		},
-		{
-			Name:  "CRYOSTAT_EXT_LISTEN_PORT",
-			Value: "443",
-		},
-		{
-			Name:  "CRYOSTAT_LISTEN_HOST",
-			Value: "cryostat-command.example.com",
-		},
 		{
 			Name:  "CRYOSTAT_CONFIG_PATH",
 			Value: "/opt/cryostat.d/conf.d",
@@ -768,17 +1203,70 @@ func NewCoreEnvironmentVariables(minimal bool, tls bool, openshift bool) []corev
 			Name:  "CRYOSTAT_CLIENTLIB_PATH",
 			Value: "/opt/cryostat.d/clientlib.d",
 		},
+		{
+			Name:  "CRYOSTAT_PROBE_TEMPLATE_PATH",
+			Value: "/opt/cryostat.d/probes.d",
+		},
+		{
+			Name:  "CRYOSTAT_ENABLE_JDP_BROADCAST",
+			Value: "false",
+		},
+		{
+			Name:  "CRYOSTAT_TARGET_CACHE_SIZE",
+			Value: "-1",
+		},
+		{
+			Name:  "CRYOSTAT_TARGET_CACHE_TTL",
+			Value: "10",
+		},
 	}
-	if !minimal {
+
+	if externalTLS {
 		envs = append(envs,
 			corev1.EnvVar{
-				Name:  "GRAFANA_DASHBOARD_URL",
-				Value: "https://cryostat-grafana.example.com",
-			},
+				Name:  "CRYOSTAT_EXT_WEB_PORT",
+				Value: "443",
+			})
+	} else {
+		envs = append(envs,
+			corev1.EnvVar{
+				Name:  "CRYOSTAT_EXT_WEB_PORT",
+				Value: "80",
+			})
+	}
+
+	if !minimal {
+		envs = append(envs,
 			corev1.EnvVar{
 				Name:  "GRAFANA_DATASOURCE_URL",
 				Value: "http://127.0.0.1:8080",
 			})
+		if externalTLS {
+			envs = append(envs,
+				corev1.EnvVar{
+					Name:  "GRAFANA_DASHBOARD_EXT_URL",
+					Value: "https://cryostat-grafana.example.com",
+				})
+		} else {
+			envs = append(envs,
+				corev1.EnvVar{
+					Name:  "GRAFANA_DASHBOARD_EXT_URL",
+					Value: "http://cryostat-grafana.example.com",
+				})
+		}
+		if tls {
+			envs = append(envs,
+				corev1.EnvVar{
+					Name:  "GRAFANA_DASHBOARD_URL",
+					Value: "https://cryostat-health.local:3000",
+				})
+		} else {
+			envs = append(envs,
+				corev1.EnvVar{
+					Name:  "GRAFANA_DASHBOARD_URL",
+					Value: "http://cryostat-health.local:3000",
+				})
+		}
 	}
 	if !tls {
 		envs = append(envs,
@@ -786,6 +1274,13 @@ func NewCoreEnvironmentVariables(minimal bool, tls bool, openshift bool) []corev
 				Name:  "CRYOSTAT_DISABLE_SSL",
 				Value: "true",
 			})
+		if externalTLS {
+			envs = append(envs,
+				corev1.EnvVar{
+					Name:  "CRYOSTAT_SSL_PROXIED",
+					Value: "true",
+				})
+		}
 	} else {
 		envs = append(envs, corev1.EnvVar{
 			Name:  "KEYSTORE_PATH",
@@ -800,7 +1295,28 @@ func NewCoreEnvironmentVariables(minimal bool, tls bool, openshift bool) []corev
 			},
 			corev1.EnvVar{
 				Name:  "CRYOSTAT_AUTH_MANAGER",
-				Value: "io.cryostat.net.OpenShiftAuthManager",
+				Value: "io.cryostat.net.openshift.OpenShiftAuthManager",
+			},
+			corev1.EnvVar{
+				Name:  "CRYOSTAT_OAUTH_CLIENT_ID",
+				Value: "cryostat",
+			},
+			corev1.EnvVar{
+				Name:  "CRYOSTAT_OAUTH_ROLE",
+				Value: "cryostat-operator-oauth-client",
+			})
+	}
+	if reportsUrl != "" {
+		envs = append(envs,
+			corev1.EnvVar{
+				Name:  "CRYOSTAT_REPORT_GENERATOR",
+				Value: reportsUrl,
+			})
+	} else {
+		envs = append(envs,
+			corev1.EnvVar{
+				Name:  "CRYOSTAT_REPORT_GENERATION_MAX_HEAP",
+				Value: "200",
 			})
 	}
 	return envs
@@ -837,6 +1353,45 @@ func NewDatasourceEnvironmentVariables() []corev1.EnvVar {
 	}
 }
 
+func NewReportsEnvironmentVariables(tls bool, resources corev1.ResourceRequirements) []corev1.EnvVar {
+	opts := "-XX:+PrintCommandLineFlags -XX:ActiveProcessorCount=1 -Dorg.openjdk.jmc.flightrecorder.parser.singlethreaded=true"
+	if !resources.Limits.Cpu().IsZero() {
+		// Assume 2 CPU limit
+		opts = "-XX:+PrintCommandLineFlags -XX:ActiveProcessorCount=2 -Dorg.openjdk.jmc.flightrecorder.parser.singlethreaded=false"
+	}
+	envs := []corev1.EnvVar{
+		{
+			Name:  "QUARKUS_HTTP_HOST",
+			Value: "0.0.0.0",
+		},
+		{
+			Name:  "JAVA_OPTIONS",
+			Value: opts,
+		},
+	}
+	if tls {
+		envs = append(envs, corev1.EnvVar{
+			Name:  "QUARKUS_HTTP_SSL_PORT",
+			Value: "10000",
+		}, corev1.EnvVar{
+			Name:  "QUARKUS_HTTP_SSL_CERTIFICATE_KEY_FILE",
+			Value: "/var/run/secrets/operator.cryostat.io/cryostat-reports-tls/tls.key",
+		}, corev1.EnvVar{
+			Name:  "QUARKUS_HTTP_SSL_CERTIFICATE_FILE",
+			Value: "/var/run/secrets/operator.cryostat.io/cryostat-reports-tls/tls.crt",
+		}, corev1.EnvVar{
+			Name:  "QUARKUS_HTTP_INSECURE_REQUESTS",
+			Value: "disabled",
+		})
+	} else {
+		envs = append(envs, corev1.EnvVar{
+			Name:  "QUARKUS_HTTP_PORT",
+			Value: "10000",
+		})
+	}
+	return envs
+}
+
 func NewCoreEnvFromSource(tls bool) []corev1.EnvFromSource {
 	envsFrom := []corev1.EnvFromSource{
 		{
@@ -871,6 +1426,37 @@ func NewGrafanaEnvFromSource() []corev1.EnvFromSource {
 	}
 }
 
+func NewWsConnectionsEnv() []corev1.EnvVar {
+	return []corev1.EnvVar{
+		{
+			Name:  "CRYOSTAT_MAX_WS_CONNECTIONS",
+			Value: "10",
+		},
+	}
+}
+
+func NewReportSubprocessHeapEnv() []corev1.EnvVar {
+	return []corev1.EnvVar{
+		{
+			Name:  "CRYOSTAT_REPORT_GENERATION_MAX_HEAP",
+			Value: "500",
+		},
+	}
+}
+
+func NewJmxCacheOptionsEnv() []corev1.EnvVar {
+	return []corev1.EnvVar{
+		{
+			Name:  "CRYOSTAT_TARGET_CACHE_SIZE",
+			Value: "10",
+		},
+		{
+			Name:  "CRYOSTAT_TARGET_CACHE_TTL",
+			Value: "20",
+		},
+	}
+}
+
 func NewCoreVolumeMounts(tls bool) []corev1.VolumeMount {
 	mounts := []corev1.VolumeMount{
 		{
@@ -897,12 +1483,23 @@ func NewCoreVolumeMounts(tls bool) []corev1.VolumeMount {
 			MountPath: "/opt/cryostat.d/clientlib.d",
 			SubPath:   "clientlib",
 		},
+		{
+			Name:      "cryostat",
+			ReadOnly:  false,
+			MountPath: "/opt/cryostat.d/probes.d",
+			SubPath:   "probes",
+		},
 		{
 			Name:      "cryostat",
 			ReadOnly:  false,
 			MountPath: "truststore",
 			SubPath:   "truststore",
 		},
+		{
+			Name:      "cert-secrets",
+			ReadOnly:  true,
+			MountPath: "/truststore/operator",
+		},
 	}
 	if tls {
 		mounts = append(mounts,
@@ -910,11 +1507,6 @@ func NewCoreVolumeMounts(tls bool) []corev1.VolumeMount {
 				Name:      "keystore",
 				ReadOnly:  true,
 				MountPath: "/var/run/secrets/operator.cryostat.io/cryostat-tls",
-			},
-			corev1.VolumeMount{
-				Name:      "cert-secrets",
-				ReadOnly:  true,
-				MountPath: "/truststore/operator",
 			})
 	}
 	return mounts
@@ -933,6 +1525,19 @@ func NewGrafanaVolumeMounts(tls bool) []corev1.VolumeMount {
 	return mounts
 }
 
+func NewReportsVolumeMounts(tls bool) []corev1.VolumeMount {
+	mounts := []corev1.VolumeMount{}
+	if tls {
+		mounts = append(mounts,
+			corev1.VolumeMount{
+				Name:      "reports-tls-secret",
+				MountPath: "/var/run/secrets/operator.cryostat.io/cryostat-reports-tls",
+				ReadOnly:  true,
+			})
+	}
+	return mounts
+}
+
 func NewVolumeMountsWithTemplates(tls bool) []corev1.VolumeMount {
 	return append(NewCoreVolumeMounts(tls),
 		corev1.VolumeMount{
@@ -1002,11 +1607,38 @@ func NewDatasourceLivenessProbe() *corev1.Probe {
 	}
 }
 
-func NewDeploymentSelector() *metav1.LabelSelector {
+func NewReportsLivenessProbe(tls bool) *corev1.Probe {
+	protocol := corev1.URISchemeHTTPS
+	if !tls {
+		protocol = corev1.URISchemeHTTP
+	}
+	return &corev1.Probe{
+		Handler: corev1.Handler{
+			HTTPGet: &corev1.HTTPGetAction{
+				Port:   intstr.IntOrString{IntVal: 10000},
+				Path:   "/health",
+				Scheme: protocol,
+			},
+		},
+	}
+}
+
+func NewMainDeploymentSelector() *metav1.LabelSelector {
 	return &metav1.LabelSelector{
 		MatchLabels: map[string]string{
-			"app":  "cryostat",
-			"kind": "cryostat",
+			"app":       "cryostat",
+			"kind":      "cryostat",
+			"component": "cryostat",
+		},
+	}
+}
+
+func NewReportsDeploymentSelector() *metav1.LabelSelector {
+	return &metav1.LabelSelector{
+		MatchLabels: map[string]string{
+			"app":       "cryostat",
+			"kind":      "cryostat",
+			"component": "reports",
 		},
 	}
 }
@@ -1015,9 +1647,9 @@ func NewVolumes(minimal bool, tls bool) []corev1.Volume {
 	return newVolumes(minimal, tls, nil)
 }
 
-func NewVolumesWithSecrets() []corev1.Volume {
+func NewVolumesWithSecrets(tls bool) []corev1.Volume {
 	mode := int32(0440)
-	return newVolumes(false, true, []corev1.VolumeProjection{
+	return newVolumes(false, tls, []corev1.VolumeProjection{
 		{
 			Secret: &corev1.SecretProjection{
 				LocalObjectReference: corev1.LocalObjectReference{
@@ -1101,33 +1733,22 @@ func newVolumes(minimal bool, tls bool, certProjections []corev1.VolumeProjectio
 			},
 		},
 	}
+	projs := append([]corev1.VolumeProjection{}, certProjections...)
 	if tls {
-		projs := []corev1.VolumeProjection{
-			{
-				Secret: &corev1.SecretProjection{
-					LocalObjectReference: corev1.LocalObjectReference{
-						Name: "cryostat-tls",
-					},
-					Items: []corev1.KeyToPath{
-						{
-							Key:  "ca.crt",
-							Path: "cryostat-ca.crt",
-							Mode: &readOnlymode,
-						},
+		projs = append(projs, corev1.VolumeProjection{
+			Secret: &corev1.SecretProjection{
+				LocalObjectReference: corev1.LocalObjectReference{
+					Name: "cryostat-tls",
+				},
+				Items: []corev1.KeyToPath{
+					{
+						Key:  "ca.crt",
+						Path: "cryostat-ca.crt",
+						Mode: &readOnlymode,
 					},
 				},
 			},
-		}
-		projs = append(projs, certProjections...)
-		volumes = append(volumes,
-			corev1.Volume{
-				Name: "cert-secrets",
-				VolumeSource: corev1.VolumeSource{
-					Projected: &corev1.ProjectedVolumeSource{
-						Sources: projs,
-					},
-				},
-			})
+		})
 
 		volumes = append(volumes,
 			corev1.Volume{
@@ -1157,9 +1778,36 @@ func newVolumes(minimal bool, tls bool, certProjections []corev1.VolumeProjectio
 				})
 		}
 	}
+
+	volumes = append(volumes,
+		corev1.Volume{
+			Name: "cert-secrets",
+			VolumeSource: corev1.VolumeSource{
+				Projected: &corev1.ProjectedVolumeSource{
+					Sources: projs,
+				},
+			},
+		})
+
 	return volumes
 }
 
+func NewReportsVolumes(tls bool) []corev1.Volume {
+	if !tls {
+		return nil
+	}
+	return []corev1.Volume{
+		{
+			Name: "reports-tls-secret",
+			VolumeSource: corev1.VolumeSource{
+				Secret: &corev1.SecretVolumeSource{
+					SecretName: "cryostat-reports-tls",
+				},
+			},
+		},
+	}
+}
+
 func NewPodSecurityContext() *corev1.PodSecurityContext {
 	fsGroup := int64(18500)
 	return &corev1.PodSecurityContext{
@@ -1167,28 +1815,30 @@ func NewPodSecurityContext() *corev1.PodSecurityContext {
 	}
 }
 
-func NewNetworkConfigurationList() operatorv1beta1.NetworkConfigurationList {
-	coreSVC := resource_definitions.NewExporterService(NewCryostat())
-	coreIng := NewNetworkConfiguration(coreSVC.Name, coreSVC.Spec.Ports[0].Port)
-
-	commandSVC := resource_definitions.NewCommandChannelService(NewCryostat())
-	commandIng := NewNetworkConfiguration(commandSVC.Name, commandSVC.Spec.Ports[0].Port)
+func NewNetworkConfigurationList(tls bool) operatorv1beta1.NetworkConfigurationList {
+	coreSVC := resource_definitions.NewCoreService(NewCryostat())
+	coreIng := NewNetworkConfiguration(coreSVC.Name, coreSVC.Spec.Ports[0].Port, tls)
 
 	grafanaSVC := resource_definitions.NewGrafanaService(NewCryostat())
-	grafanaIng := NewNetworkConfiguration(grafanaSVC.Name, grafanaSVC.Spec.Ports[0].Port)
+	grafanaIng := NewNetworkConfiguration(grafanaSVC.Name, grafanaSVC.Spec.Ports[0].Port, tls)
 
 	return operatorv1beta1.NetworkConfigurationList{
 		CoreConfig:    &coreIng,
-		CommandConfig: &commandIng,
 		GrafanaConfig: &grafanaIng,
 	}
 }
 
-func NewNetworkConfiguration(svcName string, svcPort int32) operatorv1beta1.NetworkConfiguration {
+func NewNetworkConfiguration(svcName string, svcPort int32, tls bool) operatorv1beta1.NetworkConfiguration {
 	pathtype := netv1.PathTypePrefix
-	host := "testing." + svcName
+	host := svcName + ".example.com"
+
+	var ingressTLS []netv1.IngressTLS
+	if tls {
+		ingressTLS = []netv1.IngressTLS{{}}
+	}
 	return operatorv1beta1.NetworkConfiguration{
 		Annotations: map[string]string{"nginx.ingress.kubernetes.io/backend-protocol": "HTTPS"},
+		Labels:      map[string]string{"my": "label"},
 		IngressSpec: &netv1.IngressSpec{
 			Rules: []netv1.IngressRule{
 				{
@@ -1213,19 +1863,62 @@ func NewNetworkConfiguration(svcName string, svcPort int32) operatorv1beta1.Netw
 					},
 				},
 			},
+			TLS: ingressTLS,
 		},
 	}
 }
 
-func NewServiceAccount() *corev1.ServiceAccount {
+func NewServiceAccount(isOpenShift bool) *corev1.ServiceAccount {
+	var annotations map[string]string
+	if isOpenShift {
+		annotations = map[string]string{
+			"serviceaccounts.openshift.io/oauth-redirectreference.route": `{"metadata":{"creationTimestamp":null},"reference":{"group":"","kind":"Route","name":"cryostat"}}`,
+		}
+	}
+
 	return &corev1.ServiceAccount{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      "cryostat",
 			Namespace: "default",
+			Labels: map[string]string{
+				"app": "cryostat",
+			},
+			Annotations: annotations,
 		},
 	}
 }
 
+func OtherServiceAccount() *corev1.ServiceAccount {
+	disable := false
+	return &corev1.ServiceAccount{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "cryostat",
+			Namespace: "default",
+			Labels: map[string]string{
+				"app":   "not-cryostat",
+				"other": "label",
+			},
+			Annotations: map[string]string{
+				"hello": "world",
+			},
+		},
+		ImagePullSecrets: []corev1.LocalObjectReference{
+			{
+				Name: "cryostat-dockercfg-abcde",
+			},
+		},
+		Secrets: []corev1.ObjectReference{
+			{
+				Name: "cryostat-dockercfg-abcde",
+			},
+			{
+				Name: "cryostat-token-abcde",
+			},
+		},
+		AutomountServiceAccountToken: &disable,
+	}
+}
+
 func NewRole() *rbacv1.Role {
 	return &rbacv1.Role{
 		ObjectMeta: metav1.ObjectMeta{
@@ -1360,3 +2053,14 @@ func NewConsoleLink() *consolev1.ConsoleLink {
 		},
 	}
 }
+
+func NewApiServer() *configv1.APIServer {
+	return &configv1.APIServer{
+		ObjectMeta: metav1.ObjectMeta{
+			Name: "cluster",
+		},
+		Spec: configv1.APIServerSpec{
+			AdditionalCORSAllowedOrigins: []string{"https://an-existing-user-specified\\.allowed\\.origin\\.com"},
+		},
+	}
+}
